# -*- coding: utf-8 -*-
"""
    DebiAI module for an easy data manipulation with the DebiAI app.

    Author : IRT-SystemX
    Contact : debiai@irt-systemx.fr
    GitHub : git@github.com:DebiAI/py-debiai.git
    Licence : Apache 2.0
"""

from typing import List

import utils as utils
from .debiai_project import Debiai_project


class Debiai:
    """
    Each Debiai object represent a Debiai server instance
    """

    debiai_url = ""

    def __init__(self, debiai_url: str):
        self.debiai_url = debiai_url

        # Check if the url is valid
        if self.debiai_url is None or self.debiai_url == "":
            raise ValueError("Backend url cannot be empty")

        # Remove trailing slash
        if self.debiai_url[-1] == "/":
            self.debiai_url = self.debiai_url[:-1]

        utils.check_back(debiai_url)

    def get_projects(self) -> List[Debiai_project]:
        """
        Return the server existing projects
        """
        projects = []
        projects_list = utils.get_projects(self.debiai_url)

        for project in projects_list:
            id = project["id"]
            name = project["name"]
            projects.append(Debiai_project(name, id, self.debiai_url))

        return projects

    def get_project(self, project_id: str) -> Debiai_project or None:
        """
        Return a project by name, returns none if the project doesn't exist
        """
        project = utils.get_project(self.debiai_url, project_id)
        if project != None:
            return Debiai_project(project["name"], project_id, self.debiai_url)
        else:
            return None

    def create_project(self, project_name: str) -> Debiai_project:
        """
        Create a new project from a name
        return the created Debiai_project object
        """
        if project_name is None or project_name == "":
            raise ValueError("Project name cannot be empty")

        project_id = utils.post_project(self.debiai_url, project_name)

        return Debiai_project(project_name, project_id, self.debiai_url)

    def delete_project(self, project: Debiai_project) -> bool:
        """
        Remove a project from the server
        """
        if project is None:
            raise ValueError("Project cannot be None")
        if type(project) is not Debiai_project:
            raise ValueError(
                "Parameter must be a Debiai_project object, not a "
                + type(project).__name__
                + ", use delete_project_byId instead"
            )

        return utils.delete_project(self.debiai_url, project.id)

    def delete_project_byId(self, projectId: str) -> bool:
        """
        Remove a project from the server
        """
        return utils.delete_project(self.debiai_url, projectId)



import hashlib
import pandas as pd
import numpy as np
from typing import List

# Models
from .debiai_model import Debiai_model
from .debiai_selection import Debiai_selection
from .debiai_tag import Debiai_tag

# Services
import utils as utils
import debiai_utils as debiai_utils
from .debiai_services.df_to_dict_tree import df_to_dict_tree
from .debiai_services.np_to_dict import check_np_array, np_to_dict

DEBIAI_TYPES = ["contexts", "inputs", "groundTruth", "others"]


class Debiai_project:
    """
    A Debiai project
    """

    def __init__(self, name: str, id: str, debiai_url: str):
        self.name = name
        self.id = id
        self.debiai_url = debiai_url

        self.block_structure = None
        self.expected_results = None

        self.models = None
        self.creation_date = None
        self.update_date = None

        self.project_infos()  # Load block_structure & expected_results

        # TODO : load creation date, datasets, etc...

    def __repr__(self):
        return (
            "DEBIAI project :  "
            + str(self.name)
            + "\n"
            + "Creation date : "
            + utils.timestamp_to_date(self.creation_date)
            + "\n"
            + "Update date : "
            + utils.timestamp_to_date(self.update_date)
            + "\n"
        )

    def project_infos(self):
        project_info = utils.get_project(self.debiai_url, self.id)
        if "blockLevelInfo" in project_info:
            self.block_structure = project_info["blockLevelInfo"]
        if "resultStructure" in project_info:
            self.expected_results = project_info["resultStructure"]
        if "models" in project_info:
            self.models = project_info["models"]
        if "creationDate" in project_info:
            self.creation_date = project_info["creationDate"]
        if "updateDate" in project_info:
            self.update_date = project_info["updateDate"]
        return project_info

    # Blocks structure
    def block_structure_defined(self):
        self.project_infos()
        if self.block_structure:
            return self.block_structure
        else:
            return False

    def get_block_structure(self):
        bs = self.block_structure_defined()
        if bs:
            return bs
        else:
            raise ValueError(
                "The "
                + str(self.name)
                + " DEBIAI project block_structure hasn't been set yet"
            )

    def set_blockstructure(self, block_structure: List[dict]) -> bool:
        """
        Add a block structure to the project
        This step is required before uploading data
        Throw error if the block structure is already created

        block_structure syntax:

        [
            {
                "name": str
                "contexts": [
                    {
                        "name": str,
                        "type": 'text', 'number', 'boolean',
                        "default"?: str, number,
                        "group"?: str
                    },
                    ...
                ],
                "inputs": [
                    {
                        "name": str,
                        "type": 'text', 'number', 'boolean',
                        "default"?: str, number
                        "group"?: str
                    },
                    ...
                ],
                "groundtruth": [
                    {
                        "name": str,
                        "type": 'text', 'number', 'boolean',
                        "default"?: str, number
                        "group"?: str
                    },
                    ...
                ],
                "results": [
                    {
                        "name": str,
                        "type": 'text', 'number', 'boolean',
                        "default"?: str, number
                        "group"?: str
                    },
                    ...
                ]
            },
            ...
        ]

        The last block will be considered the sample block, and will mark the end of the tree
        At least one block is required

        """

        # Check if blocklevel structure is already created
        proj_info = self.project_infos()
        if proj_info["blockLevelInfo"] != []:
            raise ValueError("Cannot set the blocklevel structure - already created")

        # Check that there is at least one block
        if not len(block_structure):
            raise ValueError("At least a block is required in the block structure")

        # Check that all the properties are correct
        for i, block in enumerate(block_structure):
            if "name" not in block:
                raise ValueError("The 'name' is required in the block n°" + str(i + 1))

            for type_ in block:
                if type_ not in DEBIAI_TYPES and type_ != "name":
                    print(
                        "Warning : unknown block type '"
                        + type_
                        + "'. Use those block types : "
                        + str(DEBIAI_TYPES)
                    )

            for debiai_type in DEBIAI_TYPES:
                if debiai_type in block:
                    for column in block[debiai_type]:
                        if "name" not in column:
                            raise ValueError(
                                "The name of the column is required in the '"
                                + debiai_type
                                + "' in the block '"
                                + block["name"]
                                + "'"
                            )
                        if "type" not in column:
                            raise ValueError(
                                "The type of the column is required in the '"
                                + debiai_type
                                + "' in the block '"
                                + block["name"]
                                + "'"
                            )
                        if column["type"] not in ["text", "number", "boolean"]:
                            raise ValueError(
                                "Unknown type of the column '"
                                + column["name"]
                                + "' in the block '"
                                + block["name"]
                                + "'. Use : ['text', 'number', 'boolean']"
                            )

        # Set the block_structure
        utils.add_blocklevel(self.debiai_url, self.id, block_structure)
        self.block_structure = block_structure

    # Results structure
    def expected_results_defined(self):
        self.project_infos()
        if self.expected_results:
            return self.expected_results
        else:
            return False

    def get_expected_results(self):
        rs = self.expected_results_defined()
        if rs:
            return rs
        else:
            raise ValueError(
                "The "
                + str(self.name)
                + " DEBIAI project expected_results hasn't been set yet"
            )

    def set_expected_results(self, expected_results: List[dict]) -> List[dict]:
        if self.expected_results is not None:
            raise ValueError("The project expected results have been already set")

        expResults = []

        for column in expected_results:
            if "name" not in column:
                raise ValueError("The attribute 'name' is required in each column")
            if "type" not in column:
                raise ValueError("The attribute 'type' is required in each column")

            col = [c for c in expResults if c["name"] == column["name"]]
            if len(col) > 0:
                raise ValueError("Each result name need to be unique")

            newRes = {"name": column["name"], "type": column["type"]}

            if "default" in column:
                newRes["default"] = column["default"]
                # TODO check default type same as col type

            expResults.append(newRes)

        utils.post_expected_results(self.debiai_url, self.id, expResults)
        self.expected_results = expResults

    def add_expected_result(self, column: dict) -> List[dict]:
        if self.expected_results is None:
            raise ValueError("The project does not have an expected results to update")

        if "name" not in column:
            raise ValueError(
                "The attribute 'name' is required in the new result column"
            )
        if "type" not in column:
            raise ValueError(
                "The attribute 'type' is required in the new result column"
            )
        if "default" not in column:
            raise ValueError(
                "The attribute 'default' is required in the new result column"
            )

        # TODO check default type same as col type

        col = [c for c in self.expected_results if c["name"] == column["name"]]
        if len(col) > 0:
            raise ValueError("'" + column["name"] + "' is already expected as a result")

        newRes = {
            "name": column["name"],
            "type": column["type"],
            "default": column["default"],
        }

        ret = utils.post_add_expected_results(self.debiai_url, self.id, newRes)
        self.expected_results = ret
        return ret

    def remove_expected_result(self, column: str) -> List[dict]:
        if self.expected_results is None:
            raise ValueError("The project does not have an expected results to update")

        # TODO check default type same as col type

        ret = utils.remove_expected_results(self.debiai_url, self.id, column)
        self.expected_results = ret
        return ret

    # Add samples

    def add_samples(self, samples: np.array) -> bool:
        """
        Add samples to the curent project, based on his block structure.
        Each one of the block structure elements need to be present in the samples numpy array, exept for the results one.

        Exemple :
        Simple block structure :
                                =======block_1=======   ======block_2======   ========samples=========
                                context_a, context_b,   context_c, input_d,   input_e, GDT_f, result_g

        The numpy array first row should containt the folowing labels in any order:
                                block_1, context_a, context_b, block_2, context_c, input_d, samples, input_e, GDT_f
        note that the result_g is not asked.

        If one the the required labels are missing, the samples wont be uploaded.
        Any labels that are not required will be ignored
        The folowing rows, if the types are correct, will be added to the database.
        """

        self.get_block_structure()  # Check that the block_structure has been set

        # Check that the array is correct and create a column index map
        indexMap = check_np_array(self.block_structure, samples)

        SAMPLE_CHUNK_SIZE = 5000  # Nuber of sample that will be added in one chunk
        SAMPLE_TO_UPLOAD = samples.shape[0] - 1

        p_bar = utils.progress_bar("Adding samples", SAMPLE_TO_UPLOAD)
        nb_sample_added = 0

        while nb_sample_added < SAMPLE_TO_UPLOAD:
            np_to_add = samples[
                nb_sample_added + 1 : nb_sample_added + 1 + SAMPLE_CHUNK_SIZE
            ]

            dict_to_add = np_to_dict(self.block_structure, np_to_add, indexMap)

            utils.post_add_tree(self.debiai_url, self.id, dict_to_add)

            nb_sample_added += SAMPLE_CHUNK_SIZE
            p_bar.update(min([nb_sample_added, SAMPLE_TO_UPLOAD]))

        return True

    def add_samples_pd(self, df: pd.DataFrame, get_hash=None) -> bool:
        """
        Add samples to the curent project, based on his block structure.
        Each one of the block structure elements need to be present in the samples dataframe

        Exemple :
        Simple block structure :
                                =======block_1=======   ======block_2======   ========samples=========
                                context_a, context_b,   context_c, input_d,   input_e, GDT_f, result_g

        The dataframe columns should containt the folowing labels in any order:
                                block_1, context_a, context_b, block_2, context_c, input_d, samples, input_e, GDT_f
        note that the result_g is not asked.

        If one the the required labels are missing, the samples wont be uploaded.
        Any labels that aren't required will be ignored
        """
        self.get_block_structure()  # Check that the block_structure has been set

        SAMPLE_CHUNK_SIZE = 5000  # Nuber of sample that will be added in one chunk
        SAMPLE_TO_UPLOAD = df.shape[0]
        p_bar = utils.progress_bar("Adding samples", SAMPLE_TO_UPLOAD)

        nb_sample_added = 0

        while nb_sample_added < SAMPLE_TO_UPLOAD:
            df_to_add = df[nb_sample_added : nb_sample_added + SAMPLE_CHUNK_SIZE]
            dict_to_add = df_to_dict_tree(df_to_add, self.block_structure)

            utils.post_add_tree(self.debiai_url, self.id, dict_to_add)
            nb_sample_added += SAMPLE_CHUNK_SIZE
            p_bar.update(min([nb_sample_added, SAMPLE_TO_UPLOAD]))

        return True

    # Models

    def get_models(self) -> List[Debiai_model]:
        self.project_infos()
        if self.models:
            return self.models
        else:
            return []

    def get_model(self, model_name: str) -> Debiai_model or None:
        self.project_infos()
        for model in self.models:
            id = model["id"]
            name = model["name"]
            if name == model_name:
                return Debiai_model(self, id, name)
        return None

    def create_model(self, name: str, metadata: dict = {}) -> Debiai_model:
        #  check parameters
        if not name:
            raise ValueError("Can't create the model: The model name is required")

        # Call the backend
        if utils.post_model(self.debiai_url, self.id, name, metadata):
            return Debiai_model(self, name, name)
        else:
            return False

    def delete_model(self, model_name: str) -> bool:
        #  check parameters
        if not model_name:
            raise ValueError("Can't delete the model: The model name is required")
        # Find the model ID
        model = self.get_model(model_name)
        if not model:
            raise ValueError("The model '" + model_name + "' does not exist")

        # Call the backend
        utils.delete_model(self.debiai_url, self.id, model.id)

    # Hash

    def check_hash(self, hash_list: list) -> list:
        """Check list of hashs with backend"""
        res = utils.check_hash_exist(self.debiai_url, self.id, hash_list)
        return res

    def __get_hash_from_df(self, block_name: list, row, map_id: str):
        """Subfunction creating a path from a row of df and hashing it"""
        path = ""

        for name in block_name:
            if name == map_id:
                path += str(row.name)
            else:
                path += str(row[name])
            path += "/"

        hash = hashlib.sha256(path.encode("utf-8")).hexdigest()

        return hash

    def create_hash(self, df: pd.DataFrame, map_id: str = None) -> pd.DataFrame:
        """
        Create a hash column into the df
        """
        # Get block names
        block_name = []

        for block in self.block_structure:
            block_name.append(block["name"])

        # Create path to hash for each row

        df["hash"] = df.apply(
            lambda row: self.__get_hash_from_df(block_name, row, map_id), axis=1
        )

        return df

    # Selections

    def get_selections(self) -> List[Debiai_selection]:
        """
        Get from the backend the list of selections, convert it in objects and returns it
        """
        selections_json = utils.get_selections(self.debiai_url, self.id)

        selections = []
        for s in selections_json:
            selections.append(
                Debiai_selection(
                    self,
                    s["name"],
                    s["id"],
                    s["creationDate"],
                    s["nbSamples"],
                    s.get("requestId", None),
                )
            )
        return selections

    def get_selection(self, selection_name: str) -> Debiai_selection or None:
        selections = self.get_selections()
        for selection in selections:
            if selection.name == selection_name:
                return selection
        return None

    # Tags

    def get_tags(self) -> List[Debiai_tag]:
        """
        Get from the backend the list of tags, convert it in objects and returns it
        """
        tags_json = utils.get_tags(self.debiai_url, self.id)

        # Convert each request into a debiai_selection object
        tags = []
        for t in tags_json:
            tags.append(
                Debiai_tag(self, t["id"], t["name"], t["creationDate"], t["updateDate"])
            )
        return tags

    def get_tag(self, tag_name: str) -> Debiai_tag or None:
        """
        Get from the backend the list of tags,
        returns the tag with the given name or none
        """
        tags = self.get_tags()

        for t in tags:
            if t.name == tag_name:
                return t

        return None

    # Pull data

    def get_numpy(self) -> np.array:
        self.get_block_structure()  # Check that the block_structure has been set

        # Pulls all the data
        sample_tree = utils.get_project_samples(self.debiai_url, self.id)
        # print(sample_tree)
        # Create the first row with the column names
        columns = np.array([])
        for block in self.block_structure:
            columns = np.append(columns, block["name"])
            for debiai_type in DEBIAI_TYPES:
                if debiai_type in block:
                    for column in block[debiai_type]:
                        columns = np.append(columns, column["name"])

        data = debiai_utils.tree_to_array(self.block_structure, sample_tree)
        return np.vstack([columns, data])

    def get_dataframe(self) -> pd.DataFrame:
        # Pull the selected samples from the backend
        # returns a pd.DataFrame
        numpy = self.get_numpy()
        col = numpy[0]
        df = pd.DataFrame(data=numpy[1:], columns=col)

        # Convert object columns to number columns
        cols = df.columns[df.dtypes.eq("object")]
        df[cols] = df[cols].apply(pd.to_numeric, errors="ignore")

        return df

    def get_tf_dataset(self) -> "tf.data.Dataset":
        import tensorflow as tf

        self.get_block_structure()  # Check that the block_structure has been set

        excepted_inputs = []
        excepted_gdt = []

        for level in self.block_structure:
            if "inputs" in level:
                excepted_inputs += level["inputs"]
            if "groundTruth" in level:
                excepted_gdt += level["groundTruth"]

        return tf.data.Dataset.from_generator(
            self.__load_samples,
            (tf.float32, tf.int32),
            ((len(excepted_inputs),), (len(excepted_gdt),)),
        )

    def __load_samples(self):
        PACH_SIZE = 1000  # Pull samples each PACH_SIZE samples

        i = 0
        while True:
            # Pull a sample tree
            sample_tree = utils.get_project_training_samples(
                self.debiai_url, self.id, i, PACH_SIZE
            )

            # Extract inputs & gdt
            inputs, gdt = debiai_utils.get_inputs_and_gdt_patch(
                self.block_structure, sample_tree
            )

            if len(inputs) == 0:
                break

            # Yield each one of the samples to the dataset
            for j in range(len(inputs)):
                yield inputs[j], gdt[j]

            # TODO : clean progress bar
            i += PACH_SIZE

    def get_tf_dataset_with_provided_inputs(
        self,
        input_function: "function",
        output_types: tuple,
        output_shapes: tuple,
        classes: list,
    ) -> "tf.data.Dataset":
        import tensorflow as tf

        self.get_block_structure()  # Check that the block_structure has been set

        self.dataset_generator_input_function = input_function
        self.dataset_generator_classes = classes

        return tf.data.Dataset.from_generator(
            self.__load_samples_with_provided_inputs,
            output_types=output_types,
            output_shapes=output_shapes,
        )

    def __load_samples_with_provided_inputs(self):
        PACH_SIZE = 1000  # Pull samples each PACH_SIZE samples

        # Only deal with 1 gdt TODO : deal with the case with more than 1 gdt

        i = 0
        while True:
            # Pull a sample tree
            sample_tree = utils.get_project_training_samples(
                self.debiai_url, self.id, i, PACH_SIZE
            )

            # Extract samples & gdt
            samples, gdt = debiai_utils.get_samples_and_gdt_patch(
                self.block_structure, sample_tree
            )

            if len(samples) == 0:
                break

            # Yield each one of the samples to the dataset
            for j in range(len(samples)):
                inputs = self.dataset_generator_input_function(samples[j])
                gdt_number = self.dataset_generator_classes.index(gdt[j][0])
                yield inputs, [gdt_number]

            i += PACH_SIZE

            # TODO : create a clean progress bar



import pytest
from debiai.debiai import Debiai
from debiai.debiai_project import Debiai_project
from debiai.config import get_config

config = get_config()
debiai_instance = Debiai(config.debiai_app_url)

print(
    "\n\nTesting the DebiAI Python module on the", config.debiai_app_url, "DebiAI app\n"
)

PROJECT_NAME = "test_project"


def test_project_list():
    projects = debiai_instance.get_projects()
    assert isinstance(projects, list)


def test_project_creation():
    nb_projects = len(debiai_instance.get_projects())

    # Create a project
    project = debiai_instance.create_project(PROJECT_NAME)

    assert isinstance(project, Debiai_project)
    assert len(debiai_instance.get_projects()) == nb_projects + 1

    # Test few edge cases
    with pytest.raises(ValueError) as excinfo:
        debiai_instance.create_project(None)
    assert "cannot be empty" in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        debiai_instance.create_project("")
    assert "cannot be empty" in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        debiai_instance.create_project(PROJECT_NAME)
    assert "already exist" in str(excinfo.value)

    # Get project by name
    project = debiai_instance.get_project(PROJECT_NAME)

    assert project is not None
    assert project.name == PROJECT_NAME

    # delete project
    assert debiai_instance.delete_project(project)
    assert len(debiai_instance.get_projects()) == nb_projects
    assert debiai_instance.get_project(PROJECT_NAME) is None


def test_project_deletion_by_id():
    project = debiai_instance.create_project(PROJECT_NAME)
    nb_projects = len(debiai_instance.get_projects())
    debiai_instance.delete_project_byId(project.id)

    assert len(debiai_instance.get_projects()) == nb_projects - 1



from typing import List
import numpy as np
import pandas as pd
import hashlib

import utils as utils
import debiai_utils as debiai_utils

DEBIAI_TYPES = ["contexts", "inputs", "groundTruth", "others"]


class Debiai_tag:
    """
    A Debiai tag : a list of sample id and a tag value
    """

    def __init__(
        self, project: str, id: str, name: str, creation_date: int, update_date: int
    ):
        # Only the overvew is loaded at init
        self.project = project
        self.id = id
        self.name = name
        self.creation_date = creation_date
        self.update_date = update_date
        self.tags = None

    def load_tags(self):
        if self.tags is None:
            # TODO: check if version has changed before loading
            tag = utils.get_tag(self.project.debiai_url, self.project.id, self.id)

            if tag is None or "tags" not in tag or tag["tags"] is None:
                raise ValueError("Error while loading the tag values")

            self.tags = tag["tags"]

    @property
    def tag_number(self) -> dict:
        # Return the number each tag is present in the tag list
        self.load_tags()

        values = {}

        for sampleHash in self.tags.keys():
            if self.tags[sampleHash] in values:
                values[self.tags[sampleHash]] += 1
            else:
                values[self.tags[sampleHash]] = 1

        return values

    def __repr__(self):
        return (
            "DEBIAI tag : '" + str(self.name) + "'\n"
            "creation date : '" + utils.timestamp_to_date(self.creation_date) + "'\n"
            "update date : '" + utils.timestamp_to_date(self.update_date) + "'\n"
        )

    def get_numpy(self, tag_value: int) -> np.array:
        # Pull all samples with the given tag value
        sampleTree = utils.get_samples_from_tag(
            self.project.debiai_url, self.project.id, self.id, tag_value
        )

        block_structure = self.project.project_infos()["blockLevelInfo"]

        # TODO : regroup in one service

        columns = np.array([])
        # Create the first row with the column names
        for block in block_structure:
            columns = np.append(columns, block["name"])
            for debiai_type in DEBIAI_TYPES:
                if debiai_type in block:
                    for column in block[debiai_type]:
                        columns = np.append(columns, column["name"])

        data = debiai_utils.tree_to_array(block_structure, sampleTree)
        if len(data) == 0:
            return np.array([columns])
        return np.vstack([columns, data])

    def get_dataframe(self, tag_value: int) -> pd.DataFrame:
        # Pull the tagged samples from the backend
        # returns a pd.DataFrame
        numpy = self.get_numpy(tag_value)
        col = numpy[0]
        df = pd.DataFrame(data=numpy[1:], columns=col)

        # Convert object columns to number columns
        cols = df.columns[df.dtypes.eq("object")]
        df[cols] = df[cols].apply(pd.to_numeric, errors="ignore")

        return df



# -*- coding: utf-8 -*-
"""
    Utils : request manager
"""
# IMPORT
import sys
import requests
import logging
import json
from typing import List
import time
import math

# GLOBAL VARIABLES
logging.basicConfig(filename="debiai.log", filemode="w", level=logging.INFO)

PYTHON_DATA_PROVIDER_ID = "Python module Data Provider"

CONNECTION_ERROR_MESSAGE = "Unable to connect to the DebiAI backend at the url : "
APPLICATION_ERROR_MESSAGE = "Something went wrong the DebiAI, is your module version the same as the application ?"


# Progress bar
class progress_bar:
    BAR_SIZE = 40

    def __init__(self, name: str, size: int, para: str = ""):
        self.name = name
        self.size = size
        self.para = para

        self.start = time.time()
        self.update(0)

    def update(self, current_progression: int):
        if self.size == 0:
            sys.stdout.write(self.name + " : Progression bar size is 0")
            return

        percent = 100.0 * current_progression / self.size
        sys.stdout.write("\r")
        sys.stdout.write(
            self.name
            + " : [{:{}}] {:>3}%".format(
                "=" * int(percent / (100.0 / self.BAR_SIZE)),
                self.BAR_SIZE,
                int(percent),
            )
        )
        sys.stdout.write(" " + str(current_progression) + "/" + str(self.size))
        sys.stdout.write(" " + str(self.para))
        sys.stdout.write(" " + str(math.floor(time.time() - self.start)) + "s")
        sys.stdout.flush()

        if percent == 100:
            print("")


# Dates
def timestamp_to_date(timestamp):
    """Convert timestamp to date"""
    return str(time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(timestamp / 1000)))


# Connection
def check_back(debiai_url):
    """Check the connection with backend"""
    try:
        ret = requests.get(debiai_url + "version")

        if "Online" not in ret.text:
            try:
                ret2 = requests.get(debiai_url)
                if "Online" not in ret2.text:
                    raise ConnectionError(
                        "An application is running on the url : "
                        + debiai_url
                        + " but this is not DebiAI"
                    )
                logging.info("DebiAI Server is up !\n")
                return True
            except (
                requests.exceptions.ConnectionError,
                requests.exceptions.Timeout,
                requests.exceptions.RequestException,
            ):
                logging.warning("Backend is down")
                raise ConnectionError(CONNECTION_ERROR_MESSAGE + debiai_url)

        logging.info("DebiAI Server is up !\n")
        return True
    except (
        requests.exceptions.ConnectionError,
        requests.exceptions.Timeout,
        requests.exceptions.RequestException,
    ):
        logging.warning("Backend is down")
        raise ConnectionError(
            "Unable to connect to the DebiAI backend at the url : " + debiai_url
        )


def projects_url(debiai_url):
    return debiai_url + "/data-providers/" + PYTHON_DATA_PROVIDER_ID + "/projects"


def project_url(debiai_url, project_id):
    return projects_url(debiai_url) + "/" + project_id


# Projects
def get_projects(debiai_url):
    """Return projects list as JSON"""
    r = requests.request("GET", projects_url(debiai_url))
    logging.info("Get_projects response: " + str(r.status_code))
    logging.info(r.text)
    return json.loads(r.text)


def get_project(debiai_url, id):
    """Return project (JSON) from id"""
    r = requests.request("GET", project_url(debiai_url, id))
    logging.info("Get_project response: " + str(r.status_code))
    logging.info(r.text)
    if r.status_code == 404:
        return None
    return json.loads(r.text)


def post_project(debiai_url, name):
    """Post new project and return project id"""
    data = {"projectName": name, "blockLevelInfo": [{"name": "file"}]}
    r = requests.request("POST", url=debiai_url + "/projects", json=data)
    if r.status_code != 200:
        raise ValueError(json.loads(r.text))
    info = json.loads(r.text)
    return info["id"]


def delete_project(debiai_url, id):
    """Delete project from id"""
    try:
        r = requests.request("DELETE", url=project_url(debiai_url, id))
        if r.status_code != 200:
            raise ValueError(json.loads(r.text))

        logging.info("Deleted project: " + id)
        return True
    except requests.exceptions.RequestException:
        return False


# Block structure
def post_expected_results(debiai_url, id, expected_results):
    """set the expected_results to a project"""
    r = requests.request(
        "POST",
        url=project_url(debiai_url, id) + "/resultsStructure",
        json=expected_results,
    )
    if r.status_code != 200:
        raise ValueError(json.loads(r.text))
    return json.loads(r.content)


def add_blocklevel(debiai_url, id, blocklevel):
    """
    Add blocklevel to a project block strucure
    Not used very much, should be removed
    TODO - Check if blocklevel already exists
    """
    r = requests.request(
        "POST",
        url=project_url(debiai_url, id) + "/blocklevels",
        json=blocklevel,
    )
    logging.info("Add block response :" + str(r.status_code) + "-" + str(r.text))
    if r.status_code != 200:
        raise ValueError(json.loads(r.text))
    logging.info("Added blocklevel to project " + id)


def post_add_expected_results(debiai_url, id, expected_result):
    """Add expected_result to a project"""
    r = requests.request(
        "POST",
        url=project_url(debiai_url, id) + "/expectedResult",
        json=expected_result,
    )
    if r.status_code != 200:
        raise ValueError(json.loads(r.text))
    return json.loads(r.content)


def remove_expected_results(debiai_url, id, expected_result):
    """remove expected_result from a project"""
    obj = {"value": expected_result}

    r = requests.request(
        "POST",
        url=project_url(debiai_url, id) + "/del_expectedResult",
        json=obj,
    )
    if r.status_code != 200:
        raise ValueError(json.loads(r.text))
    return json.loads(r.content)


# Selections
def get_selections(debiai_url, id):
    """Return a project get_selections as JSON"""
    r = requests.request("GET", project_url(debiai_url, id) + "/selections")
    logging.info("get_requests response: " + str(r.status_code))
    logging.info(r.text)
    return json.loads(r.text)


# Models
def post_model(debiai_url, id, name, metadata):
    """Add to an existing project a tree of samples"""
    data = {"name": name, "metadata": metadata}

    r = requests.request("POST", url=project_url(debiai_url, id) + "/models", json=data)

    if r.status_code == 409:
        print("Warning : The model " + name + " already exists")
        return 409
    if r.status_code != 200:
        raise ValueError("post_model : " + json.loads(r.text))
    return True


def post_model_results_dict(
    debiai_url, project_id, modelId, results: dict, expected_results_order: List[str]
):
    """Add to an existing project model some results from a tree dict"""
    data = {"results": results, "expected_results_order": expected_results_order}
    try:
        r = requests.request(
            "POST",
            url=project_url(debiai_url, project_id)
            + "/models/"
            + modelId
            + "/resultsDict",
            json=data,
        )

        if r.status_code != 200:
            raise ValueError("post_model_results_dict : " + json.loads(r.text))
        return True

    except json.decoder.JSONDecodeError as e:
        raise ValueError("internal server error")


def delete_model(debiai_url, project_id, model_id):
    """Delete a model from a project"""
    try:
        r = requests.request(
            "DELETE",
            url=project_url(debiai_url, project_id) + "/models/" + model_id,
        )
        if r.status_code != 200:
            raise ValueError(json.loads(r.text))
        logging.info("Deleted model: " + model_id)
        return True
    except requests.exceptions.RequestException:
        return False


# Tags
def get_tags(debiai_url, project_id):
    """Return a tag as JSON form id"""
    r = requests.request("GET", url=project_url(debiai_url, project_id) + "/tags")
    logging.info("get_tags response: " + str(r.status_code))
    logging.info(r.text)
    return json.loads(r.text)


def get_tag(debiai_url, project_id, tag_id):
    """Return a tag as JSON form id"""
    r = requests.request(
        "GET",
        url=project_url(debiai_url, project_id) + "/tags/" + str(tag_id),
    )
    logging.info("get_tag response: " + str(r.status_code))
    logging.info(r.text)
    return json.loads(r.text)


def get_samples_from_tag(debiai_url, project_id, tag_id, tag_value):
    """Return a sample tree (JSON)"""
    r = requests.request(
        "GET",
        url=project_url(debiai_url, project_id)
        + "/tags/"
        + str(tag_id)
        + "/samples/"
        + str(tag_value),
    )
    logging.info("get_samples_from_tag response: " + str(r.status_code))
    return json.loads(r.text)


# Sample tree
def post_add_tree(debiai_url, project_id, tree):
    """
    Add to an existing project a tree of samples

    The expected tree format :
    [
        {
            "name": "b1",
            "contexts": [..],
            "inputs": [...],
            "groundTruth": [...],
            "others: [...]"
            "childrenInfoList": [
                {
                    "name": "b1-1",
                    "contexts": [...],
                    "inputs": [...],
                    "groundTruth": [...],
                    "others: [...]"
                    "childrenInfoList": [...]
                },
                ...
            ]
        },
        ...
    ]
    """

    data = {"blockTree": tree}

    r = requests.request(
        "POST",
        url=project_url(debiai_url, project_id) + "/blocks",
        json=data,
    )
    if r.status_code == 201:
        print("No block added")
    elif r.status_code != 200:
        raise ValueError("Internal server error while adding the datatree")
    return True


def get_project_samples(debiai_url, project_id, depth=0):
    """Return a sample tree (JSON)"""
    r = requests.request(
        "GET",
        url=project_url(debiai_url, project_id) + "/blocks?depth=" + str(depth),
    )
    logging.info("get_project_samples response: " + str(r.status_code))
    return json.loads(r.text)


def get_samples_from_selection(debiai_url, project_id, selectionId, depth=0):
    """Return a sample tree (JSON)"""
    r = requests.request(
        "GET",
        url=project_url(debiai_url, project_id)
        + "/blocks/"
        + selectionId
        + "?depth="
        + str(depth),
    )
    logging.info("get_samples_from_selection response: " + str(r.status_code))
    return json.loads(r.text)


def get_project_training_samples(debiai_url, project_id, start, size):
    """Return a sample inputs and gdt array ready to be processed"""
    r = requests.request(
        "GET",
        url=project_url(debiai_url, project_id)
        + "/trainingSamples?start="
        + str(start)
        + "&size="
        + str(size),
    )
    logging.info("get_project_training_samples response: " + str(r.status_code))
    return json.loads(r.text)


def get_training_samples_from_selection(
    debiai_url, project_id, selectionId, start, size
):
    """Return a sample inputs and gdt array ready to be processed"""
    r = requests.request(
        "GET",
        url=project_url(debiai_url, project_id)
        + "/trainingSamples?selectionId="
        + selectionId
        + "&start="
        + str(start)
        + "&size="
        + str(size),
    )
    logging.info("get_training_samples_from_selection response: " + str(r.status_code))
    return json.loads(r.text)


# Hash
def check_hash_exist(debiai_url, project_id, hash_list):
    """Check with backend if hashes exists"""
    data = {"hash_list": hash_list}

    try:
        r = requests.request(
            "POST",
            url=project_url(debiai_url, project_id) + "/check_hash",
            json=data,
        )
        if r.status_code != 200:
            raise ValueError("check_hash_exist : " + json.loads(r.text))
        return json.loads(r.text)
    except json.decoder.JSONDecodeError as e:
        raise ValueError("internal server error")


def post_results_hash(debiai_url, project_id, modelId, results: dict):
    """Add to an existing project model some results from a hash tree"""
    data = {
        "results": results,
    }
    try:
        r = requests.request(
            "POST",
            url=project_url(debiai_url, project_id)
            + "/models/"
            + modelId
            + "/resultsHash",
            json=data,
        )

        if r.status_code != 200:
            raise ValueError("post_model_results_dict : " + json.loads(r.text))
        return True
    except json.decoder.JSONDecodeError as e:
        raise ValueError("internal server error")



import pandas as pd

DEBIAI_TYPES = ["contexts", "inputs", "groundTruth", "others"]


def df_to_dict_tree(df: pd.DataFrame, block_structure: list):
    col_index_map = {}
    data = df.to_dict("split")
    column = data["columns"]

    # find the position of the columns of the block structure in the dataframe
    for block in block_structure:
        if block["name"] not in column:
            raise ValueError(
                "'" + block["name"] + "' is missing from the given samples"
            )

        col_index_map[block["name"]] = column.index(block["name"])

        for type_ in DEBIAI_TYPES:
            if type_ in block:
                for col in block[type_]:
                    if col["name"] not in column:
                        raise ValueError(
                            "'"
                            + col["name"]
                            + "' "
                            + type_
                            + " is missing from the given samples"
                        )

                    col_index_map[col["name"]] = column.index(col["name"])

    # Create a tree json with only the blocks names
    block_tree_dict = {"childrenInfoList": {}}
    for sample in data["data"]:
        parent = block_tree_dict
        for level, block_level in enumerate(block_structure):
            block_name = sample[col_index_map[block_level["name"]]]
            if block_name not in parent["childrenInfoList"]:
                # Add the block to the dict
                parent["childrenInfoList"][block_name] = __create_block(
                    block_level, sample, col_index_map
                )

                # if level < len(block_structure) - 1:
                # Not the sample
                parent["childrenInfoList"][block_name]["childrenInfoList"] = {}

            parent = parent["childrenInfoList"][block_name]

    # Create the block tree from the dict (temporary)
    block_tree = __block_dict_to_array_tree(-1, block_tree_dict, block_structure)
    return block_tree["childrenInfoList"]


def __create_block(blockLevel: dict, sample: list, col_index_map: dict):
    block = {"name": str(sample[col_index_map[blockLevel["name"]]])}
    for DEBIAI_type in DEBIAI_TYPES:
        if DEBIAI_type in blockLevel:
            block[DEBIAI_type] = []
            for col in blockLevel[DEBIAI_type]:
                block[DEBIAI_type].append(sample[col_index_map[col["name"]]])
    return block


def __block_dict_to_array_tree(level, block_dict, block_structure):
    if level == len(block_structure) - 1:
        # Sample level
        return block_dict

    children_info_list = []
    for child_name in block_dict["childrenInfoList"]:
        children_info_list.append(
            __block_dict_to_array_tree(
                level + 1, block_dict["childrenInfoList"][child_name], block_structure
            )
        )

    block_dict["childrenInfoList"] = children_info_list
    return block_dict



from typing import List
import numpy as np
import pandas as pd
import hashlib

import utils as utils
import debiai_utils as debiai_utils

DEBIAI_TYPES = ["contexts", "inputs", "groundTruth", "others"]


class Debiai_selection:
    """
    A Debiai data selection : a list of sample id
    It can belongs to a request
    """

    def __init__(
        self,
        project,
        name: str,
        id: str,
        creationDate: int,
        nbSamples: str,
        requestId: str,
    ):
        self.project = project
        self.name = name
        self.id = id
        self.creationDate = creationDate
        self.nbSamples = nbSamples
        self.requestId = requestId

    def __repr__(self):
        return (
            "DEBIAI selection : '" + str(self.name) + "'\n"
            "creation date : '" + utils.timestamp_to_date(self.creationDate) + "'\n"
            "number of samples  : '" + str(self.nbSamples) + "'\n"
        )

    def get_numpy(self) -> np.array:
        # Pulls all the selection data
        sampleTree = utils.get_samples_from_selection(
            self.project.debiai_url, self.project.id, self.id
        )

        block_structure = self.project.project_infos()["blockLevelInfo"]

        """
            tree structure :
            [
                {
                    'id',
                    'creationDate',
                    'groundTruth',
                    'inputs',
                    'level',
                    'metaDataList',
                    'name',
                    'parentPath',
                    'path',
                    'updateDate',
                    'version',
                    'childrenInfoList' : {

                    }
                }
            ]
        """
        columns = np.array([])
        # Create the first row with the column names
        for block in block_structure:
            columns = np.append(columns, block["name"])
            for debiai_type in DEBIAI_TYPES:
                if debiai_type in block:
                    for column in block[debiai_type]:
                        columns = np.append(columns, column["name"])

        data = debiai_utils.tree_to_array(block_structure, sampleTree)
        return np.vstack([columns, data])

    def get_dataframe(self) -> pd.DataFrame:
        # Pull the selected samples from the backend
        # returns a pd.DataFrame
        numpy = self.get_numpy()
        col = numpy[0]
        df = pd.DataFrame(data=numpy[1:], columns=col)

        # Convert object columns to number columns
        cols = df.columns[df.dtypes.eq("object")]
        df[cols] = df[cols].apply(pd.to_numeric, errors="ignore")

        return df

    # Tensorflow dataset generator

    def get_tf_dataset(self) -> "tf.data.Dataset":
        import tensorflow as tf

        block_structure = self.project.project_infos()["blockLevelInfo"]

        excepted_inputs = []
        excepted_gdt = []

        for level in block_structure:
            if "inputs" in level:
                excepted_inputs += level["inputs"]
            if "groundTruth" in level:
                excepted_gdt += level["groundTruth"]

        return tf.data.Dataset.from_generator(
            self.__load_samples,
            (tf.float32, tf.int32),
            ((len(excepted_inputs),), (len(excepted_gdt),)),
        )

    def __load_samples(self):
        PACH_SIZE = 1000  # Pull samples each PACH_SIZE samples

        for i in range(0, self.nbSamples, PACH_SIZE):
            # Pull a sample tree
            sampleTree = utils.get_training_samples_from_selection(
                self.project.debiai_url, self.project.id, self.id, i, PACH_SIZE
            )

            # Extract inputs & gdt
            inputs, gdt = debiai_utils.get_inputs_and_gdt_patch(
                self.project.block_structure, sampleTree
            )

            # Pull undirect inputs from external source (image, ...) from inputs
            # TODO : try with Faurecia
            # Yield each one of the samples to the dataset
            for j in range(len(inputs)):
                yield inputs[j], gdt[j]

            # TODO : create a clean progress bar
            print(str(i) + "/" + str(self.nbSamples))

    def get_tf_dataset_with_provided_inputs(
        self,
        input_function: "function",
        output_types: tuple,
        output_shapes: tuple,
        classes: list,
    ) -> "tf.data.Dataset":
        import tensorflow as tf

        self.dataset_generator_input_function = input_function
        self.dataset_generator_classes = classes

        block_structure = self.project.project_infos()["blockLevelInfo"]

        return tf.data.Dataset.from_generator(
            self.__load_samples_with_provided_inputs,
            output_types=output_types,
            output_shapes=output_shapes,
        )

    def __load_samples_with_provided_inputs(self):
        PACH_SIZE = 1000  # Pull samples each PACH_SIZE samples

        # Only deal with 1 gdt TODO : deal with the case with more than 1 gdt

        for i in range(0, self.nbSamples, PACH_SIZE):
            # Pull a sample tree
            sampleTree = utils.get_training_samples_from_selection(
                self.project.debiai_url, self.project.id, self.id, i, PACH_SIZE
            )

            # Extract samples & gdt
            samples, gdt = debiai_utils.get_samples_and_gdt_patch(
                self.project.block_structure, sampleTree
            )

            # Yield each one of the samples to the dataset
            for j in range(len(samples)):
                inputs = self.dataset_generator_input_function(samples[j])
                gdt_number = self.dataset_generator_classes.index(gdt[j][0])
                yield inputs, [gdt_number]

            # TODO : create a clean progress bar



"""
DebiAI python module configuration
"""

import os
from typing import Optional


class DebiaiConfig:
    """Global configuration object for Debiai"""

    def __init__(
        self,
        debiai_app_url: str = "http://localhost:3000/",
    ):
        self.debiai_app_url = debiai_app_url

    def __repr__(self):
        return f"DebiaiConfig ( " f"debiai_app_url: {self.debiai_app_url} " f")"


_DEBIAI_CONFIG: Optional[DebiaiConfig] = None


def configure_config(provided_config: Optional[DebiaiConfig] = None):
    global _DEBIAI_CONFIG

    if provided_config is not None:
        _DEBIAI_CONFIG = provided_config
    else:
        _DEBIAI_CONFIG = DebiaiConfig(
            debiai_app_url=os.environ.get("DEBIAI_APP_URL", "http://localhost:3000/")
        )


def get_config() -> DebiaiConfig:
    global _DEBIAI_CONFIG

    if _DEBIAI_CONFIG is None:
        configure_config()

    assert _DEBIAI_CONFIG is not None
    return _DEBIAI_CONFIG



import numpy as np
import pandas as pd
from debiai.debiai import Debiai
from debiai.config import get_config

config = get_config()
debiai_instance = Debiai(config.debiai_app_url)

projects = debiai_instance.get_projects()
for p in projects:
    debiai_instance.delete_project(p)

PROJECT_NAME = "test_samples"

block_structure = [
    {
        "name": "Image ID",
        "contexts": [
            {"name": "My context 1", "type": "text"},
            {"name": "My context 2", "type": "number", "group": "My group 1"},
        ],
        "groundTruth": [{"name": "My groundtruth 1", "type": "number"}],
    }
]


expected_results = [
    {"name": "Model result", "type": "number"},
    {"name": "Model confidence", "type": "number"},
    {"name": "Model error", "type": "text"},
]


def create_empty_project():
    if debiai_instance.get_project(PROJECT_NAME) is not None:
        debiai_instance.delete_project_byId(PROJECT_NAME)

    project = debiai_instance.create_project(PROJECT_NAME)
    project.set_blockstructure(block_structure)
    project.set_expected_results(expected_results)
    return project


def test_samples_df():
    project = create_empty_project()
    samples_df = pd.DataFrame(
        {
            "Image ID": ["image-1", "image-2", "image-3"],
            "My context 1": ["A", "B", "C"],
            "My context 2": [0.28, 0.388, 0.5],
            "My groundtruth 1": [8, 7, 19],
        }
    )
    assert project.add_samples_pd(samples_df)

    model_1 = project.create_model("Model 1")
    results_df = pd.DataFrame(
        {
            "Image ID": ["image-1", "image-2", "image-3"],
            "Model result": [5, 7, 19],
            "Model confidence": [0.22, 0.8, 0.9],
            "Model error": ["yes", "no", "no"],
        }
    )
    assert model_1.add_results_df(results_df)
    debiai_instance.delete_project(project)


def test_samples_np():
    project = create_empty_project()
    samples_np = np.array(
        [
            ["Image ID", "My context 1", "My context 2", "My groundtruth 1"],
            ["image-1", "D", 0.98, 1],
            ["image-2", "E", 0.97, 3],
            ["image-3", "F", 0.8, 2],
        ]
    )
    assert project.add_samples(samples_np)

    model_2 = project.create_model("Model 2")
    results_np = np.array(
        [
            ["Image ID", "Model result", "Model confidence", "Model error"],
            ["image-1", 3, 0.98, "yes"],
            ["image-2", 7, 0.97, "no"],
            ["image-3", 10, 0.8, "yes"],
        ]
    )
    assert model_2.add_results_np(results_np)
    debiai_instance.delete_project(project)


def test_samples_multi_levels():
    multi_block_structure = [
        {"name": "Dataset ID"},
        {
            "name": "Image ID",
            "contexts": [
                {"name": "Context", "type": "text"},
            ],
            "groundTruth": [{"name": "GDT", "type": "number"}],
        },
        {"name": "Obst ID"},
    ]

    if debiai_instance.get_project("test Multi Levels") is not None:
        debiai_instance.delete_project_byId("test Multi Levels")

    project = debiai_instance.create_project("test Multi Levels")
    project.set_blockstructure(multi_block_structure)

    samples = np.array(
        [
            ["Dataset ID", "Image ID", "Context", "GDT", "Obst ID"],
            ["A", "image-1", "D", 1, 3000],
            ["A", "image-2", "E", 3, 2000],
            ["A", "image-2", "F", 2, 1000],
            ["A", "image-2", "F", 2, 1000],
        ]
    )

    assert project.add_samples(samples)
    debiai_instance.delete_project(project)



import numpy as np

DEBIAI_TYPES = ["contexts", "inputs", "groundTruth", "others"]


def np_to_dict(block_structure: list, samples: np.array, indexMap: dict):
    ret = []

    for sample in samples:
        parent = None
        #  First block exception
        if len(block_structure) > 1:
            for i, block in enumerate(ret):
                if block["name"] == sample[indexMap[block_structure[0]["name"]]]:
                    # Block already created
                    parent = ret[i]
                    continue

        if parent is None:
            #  Block not created, creating block
            newBlock = __create_block(block_structure[0], sample, indexMap)
            ret.append(newBlock)
            parent = newBlock

        #  Deal with other blocks
        for blockStruct in block_structure[1:]:
            blockNameIndex = indexMap[blockStruct["name"]]

            nextParent = None
            for i, block in enumerate(parent["childrenInfoList"]):
                if block["name"] == sample[blockNameIndex]:
                    # Block already created
                    nextParent = block
                    parent = nextParent
                    continue

            if nextParent is None:
                #  Block not created, creating block
                newBlock = __create_block(blockStruct, sample, indexMap)
                parent["childrenInfoList"].append(newBlock)
                parent = newBlock

    return ret


def check_np_array(block_structure: list, samples: np.array):
    indexMap = {}  #  map of the user structure position and there index in the samples

    # Check array compliance
    for block in block_structure:
        if block["name"] not in samples[0]:
            raise ValueError(
                "'" + block["name"] + "' is missing from the given samples"
            )

        indexMap[block["name"]] = np.where(samples[0] == block["name"])[0][0]

        for type_ in DEBIAI_TYPES:
            if type_ in block:
                for col in block[type_]:
                    if col["name"] not in samples[0]:
                        raise ValueError(
                            "'"
                            + col["name"]
                            + "' "
                            + type_
                            + " is missing from the given samples"
                        )

                    indexMap[col["name"]] = np.where(samples[0] == col["name"])[0][0]
                    #  TODO check column type
    return indexMap


def __create_block(blockStruct, sample, indexMap):
    newBlock = {
        "name": str(sample[indexMap[blockStruct["name"]]]),
        "childrenInfoList": [],
    }

    for type_ in DEBIAI_TYPES:
        if type_ in blockStruct:
            newBlock[type_] = []
            for col in blockStruct[type_]:
                colIndex = indexMap[col["name"]]
                if col["type"] == "text":
                    newBlock[type_].append(sample[colIndex])
                else:
                    newBlock[type_].append(float(sample[colIndex]))
    return newBlock



from typing import List
import numpy as np
import pandas as pd
import hashlib

import utils as utils


class Debiai_model:
    """
    A Debiai project model
    """

    def __init__(self, project: "Debiai_project", name: str, id: str):
        self.project = project
        self.name = name
        self.id = id

    def add_results_dict(
        self, results: dict, expected_results_order: List[str] = None
    ) -> bool:
        """
        Add a result to the model from a dict.
        The tree allow the user to specify to what sample he created the result
        Each one of the block results structure elements need to be present at
        the end of the tree and in the expected_results_order array.
        """
        # Update the block structure & expected results
        self.project.project_infos()

        if self.project.expected_results is None:
            raise ValueError(
                "The project expected results need to be specified before adding results"
            )

        if type(results) != dict:
            raise ValueError("Results must be of type dict")

        if expected_results_order is not None:
            for expected_result in self.project.expected_results:
                if expected_result["name"] not in expected_results_order:
                    raise ValueError(
                        "The expected result '"
                        + expected_result["name"]
                        + "' is missing from the expected_results_order Array"
                    )

            for given_expected_result in expected_results_order:
                result_expected = False
                for expected_result in self.project.expected_results:
                    if given_expected_result == expected_result["name"]:
                        result_expected = True
                if not result_expected:
                    raise ValueError(
                        "The given expected result '"
                        + given_expected_result
                        + "' is not an expected result"
                    )
        else:
            expected_results_order = list(
                map(lambda r: r["name"], self.project.expected_results)
            )

        # Start the results verification
        sampleIndex = len(self.project.block_structure) - 1

        for rootBlock in results:
            self.__checkResultDict(
                results[rootBlock],
                0,
                sampleIndex,
                expected_results_order,
                str(rootBlock),
            )

        # Upload the results
        return utils.post_model_results_dict(
            self.project.debiai_url,
            self.project.id,
            self.id,
            results,
            expected_results_order,
        )

    def add_results_df(self, results: pd.DataFrame, map_id=None) -> bool:
        """
        Add results from a dataFrame.
        """
        # Add results with __add_results_pd sequentialy
        p_bar = utils.progress_bar("Adding results", results.shape[0], self.name)
        results_added = 0

        chunk_size = 5000
        for start in range(0, results.shape[0], chunk_size):
            results_subset = results.iloc[start : start + chunk_size]

            self.__add_results_pd(results_subset, map_id=map_id)

            results_added = results_added + results_subset.shape[0]
            p_bar.update(results_added)

        return True

    def __add_results_pd(self, results: pd.DataFrame, map_id=None) -> bool:
        """Add results to the model throught DataFrame"""
        if self.project.expected_results is None:
            raise ValueError(
                "The project expected results need to be specified before adding results"
            )

        # Check if block names are in df columns.
        for block in self.project.block_structure:
            if block["name"] not in results.columns and block["name"] != map_id:
                print("'" + block["name"] + "' is missing from the given samples")
                return False

        # Dataframe form
        # b1    b2    sample    res1  res2  res3
        # b1-1  b2-1  sample-1  1     2     "a"
        # ...   ...   ...       ...   ...   ...

        # Dict form :
        # {
        #     "b1-1": {
        #         "b2-1": {
        #             "sample-1": [1, 2, "a"],
        #             ...
        #         },
        #         ...
        #     },
        #     ...
        # }

        # Exctract results name
        results_name = []
        for resu in self.project.expected_results:
            results_name.append(resu["name"])

        # Transform DataFrame into dic recursively
        dic_res = self.__pd_to_dict_recur(results, 0, results_name)

        return self.add_results_dict(dic_res)

    def __pd_to_dict_recur(self, df: pd.DataFrame, level: int, results_name) -> dict:
        blockLevel = self.project.block_structure[level]
        dictRet = {}

        if level == len(self.project.block_structure) - 1:
            # Sample level
            # Exctract results

            df_results_dict = df.to_dict("records")

            # df_results_dict records :
            # [
            #     {"sampleId": 1, "res1": 1, "res2": 1, ...},
            #     ...
            # ]

            for record in df_results_dict:
                dictRet[record[blockLevel["name"]]] = []
                for resName in results_name:
                    dictRet[record[blockLevel["name"]]].append(record[resName])

        else:
            # Other levels
            dfs = dict(tuple(df.groupby(blockLevel["name"])))

            for key in dfs:
                dictRet[key] = self.__pd_to_dict_recur(
                    dfs[key], level + 1, results_name
                )

        return dictRet

    def add_results_np(self, results: np.array) -> bool:
        """
        Add a result to the model from a np array.
        simply convert it into a df and use the already coded df import function
        """
        col = results[0]
        df = pd.DataFrame(data=results[1:], columns=col)
        self.add_results_df(df)
        return True

    def __checkResultDict(
        self,
        block,
        level: int,
        sampleIndex: int,
        expected_results_order: List[str],
        path: str,
    ):
        """
        Check recursively if a result dict is coherent with the block_structure
        """

        if level == sampleIndex:
            #  Sample level : the results : they need to be verified
            if len(block) != len(expected_results_order):
                raise ValueError(
                    "in : "
                    + path
                    + ", "
                    + str(len(block))
                    + " value where given but "
                    + str(len(expected_results_order))
                    + "where expected"
                )

        else:
            for subBlockKey in block:
                self.__checkResultDict(
                    block[subBlockKey],
                    level + 1,
                    sampleIndex,
                    expected_results_order,
                    path + " / " + str(subBlockKey),
                )

    def add_results_hash(
        self, results: dict, expected_results_order: List[str] = []
    ) -> dict:
        """
        Add results from a dict oh hash:results
        """
        if self.project.expected_results is None:
            raise ValueError(
                "The project expected results need to be specified before adding results"
            )

        # Get wrong hash
        hash_list = []

        for key in results:
            hash_list.append(key)

        wrong_hash = utils.check_hash_exist(
            self.project.debiai_url, self.project.id, hash_list
        )

        # Add expected results order
        if expected_results_order != []:
            for dic in results:
                new_results = []
                for block in self.project.expected_results:
                    new_results.append(
                        results[dic][expected_results_order.index(block["name"])]
                    )
                results[dic] = new_results

        # Remove wrong hash from the dico and add them to false dic to return
        wrong_values = {}

        for key in wrong_hash:
            wrong_values[key] = results.pop(key)

        # Push new results in backend
        err = utils.post_results_hash(
            self.project.debiai_url, self.project.id, self.id, results
        )

        return wrong_values



import numpy as np

"""
    debiai_utils : common function for all debiai classes
"""

DEBIAI_TYPES = ["contexts", "inputs", "groundTruth", "others"]


def tree_to_array(block_structure, sample_tree):
    """
    Convert a sample tree to a 2d array
    """

    if len(sample_tree) == 0:
        return []

    # Add recursively each one of the blocks
    PATCH_SIZE = 700  # Regroup the blocks after each PATCH_SIZE block
    i = 0

    ret = []
    ret_tmp = get_block_data(block_structure, 0, sample_tree[0])
    for block in sample_tree[1:]:
        i = i + 1
        if i == PATCH_SIZE:
            ret.append(ret_tmp)
            ret_tmp = get_block_data(block_structure, 0, block)
            i = 0
            continue

        ret_tmp = np.vstack([ret_tmp, get_block_data(block_structure, 0, block)])

    ret.append(ret_tmp)
    ret_final = ret[0]
    for r in ret[1:]:
        ret_final = np.vstack([ret_final, r])

    return ret_final


def get_block_data(block_structure, level, block: object):
    """
    Used by tree_to_array, returns recursively a 2d array with
    a block and his childrens data
    """
    # Add the block ID
    block_line = np.array([block["name"]])

    # Add the gdt, inputs, ...
    for debiai_type in DEBIAI_TYPES:
        if debiai_type in block_structure[level]:
            for i in range(len(block_structure[level][debiai_type])):
                block_line = np.append(block_line, block[debiai_type][i])

    # Load the childrens blocks data
    if "childrenInfoList" in block and len(block["childrenInfoList"]) > 0:
        childs_data = np.array([])
        for child_block in block["childrenInfoList"]:
            child_data = get_block_data(block_structure, level + 1, child_block)

            if childs_data.size == 0:
                childs_data = np.array(child_data)
            else:
                childs_data = np.vstack([childs_data, child_data])

        # Merge block data with childrens
        block_line = np.repeat([block_line], repeats=np.shape(childs_data)[0], axis=0)

        block_line = np.concatenate((block_line, childs_data), axis=1)
        return block_line
    else:
        return [block_line]


def get_inputs_and_gdt_patch(block_structure, sample_tree):
    """
    return from a sample tree a list of inputs and gdt
    used to create tf datasets
    """
    data = tree_to_array(block_structure, sample_tree)

    # Filter the array to keep the inputs & gdt
    inputs = []
    gdt = []

    # TODO can be done faster and cleaner
    for d in data:
        tmp_inputs = []
        tmp_gdt = []
        ind = -1
        for block in block_structure:
            ind += 1
            for debiai_type in DEBIAI_TYPES:
                if debiai_type in block:
                    for column in block[debiai_type]:
                        ind += 1
                        if debiai_type == "inputs":
                            tmp_inputs.append(float(d[ind]))
                        elif debiai_type == "groundTruth":
                            tmp_gdt.append(float(d[ind]))
        inputs.append(tmp_inputs)
        gdt.append(tmp_gdt)

    return inputs, gdt


def get_samples_and_gdt_patch(block_structure, sample_tree):
    """
    return from a sample tree a list of samples and gdt
    used to create tf datasets
    """

    data = tree_to_array(block_structure, sample_tree)

    # Filter the array to keep the samples & gdt
    samples = []
    gdt = []

    # TODO can be done faster and cleaner
    for d in data:
        tmp_samples = {}
        tmp_gdt = []
        ind = -1
        for block in block_structure:
            ind += 1
            tmp_samples[block["name"]] = d[ind]
            for debiai_type in DEBIAI_TYPES:
                if debiai_type in block:
                    for column in block[debiai_type]:
                        ind += 1
                        if column["type"] == "text":
                            tmp_samples[column["name"]] = d[ind]
                        else:
                            tmp_samples[column["name"]] = float(d[ind])

                        if debiai_type == "groundTruth":
                            if column["type"] == "text":
                                tmp_gdt.append(d[ind])
                            else:
                                tmp_gdt.append(float(d[ind]))
        samples.append(tmp_samples)
        gdt.append(tmp_gdt)

    return samples, gdt



import setuptools

setuptools.setup(
    name="debiai",
    version="0.24.0",
    author="IRT-SystemX",
    author_email="debiai@irt-systemx.fr",
    description="DebiAI python module",
    license="Apache 2.0",
    keywords="DebiAI, Data vis, AI, Bias",
    url="https://github.com/debiai/py-debiai",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
    install_requires=["numpy", "pandas", "requests"],
)
